// -*- mode: objc; tab-width: 4 -*-

//  This file is part of Byeoroo, a Korean input method server for X11.
//  Copyright (C) 1997  Park, Jae-hyon.
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License , or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//  
//  You may contact the author by:
//     e-mail:  jhpark@entropy.kaist.ac.kr
//       mail:  Park, Jae-hyon
//              Physics Department, 305-701
//              Korea Advanced Institute of Science and Technology
//              Taejon, South Korea


#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <ctype.h>
#include <dirent.h>
#include <string.h>

#import  "ChineseDic.h"
#include "sutil.h"

typedef int (*QsortComp)(const void *, const void *);
typedef QsortComp BsearchComp;

@implementation ChineseDic

- init
{
	return [self initFromSearchPath:"."];
}

- initFromSearchPath:(char *)search_path
{
	char *p, *q;
	char directory[FILENAME_MAX];
	
	if (!(self = [super init]))
		return nil;
	
	for (ndics = 0, p = search_path; *p; p = (*q ? q + 1 : q)) {
		DIR *ds;
		struct dirent *e;

		if ((q = strchr(p, ':')) == NULL)
			q = p + strlen(p);
		strmemcpy(directory, p, q - p);
		if ((ds = opendir(directory)) == NULL) {
			perror(directory);
			continue;
		}
		while ((e = readdir(ds)) != NULL) {
			if (has_extension(e->d_name, CDICEXT)) {
				PerFileDic *new_dic;
				PerFileDic **new_dic_list;
				
				strcat(strcat(directory, "/"), e->d_name);
				if ((new_dic = [[PerFileDic alloc] initFromFile:directory])) {
					new_dic_list =
						realloc(dic_list, (ndics+1) * sizeof(dic_list[0]));
					if (new_dic_list == NULL) {
						[new_dic free];
						goto ret;
					}
					dic_list = new_dic_list;
					dic_list[ndics++] = new_dic;
				}
				directory[q - p] = '\0';
			}
		}
		closedir(ds);
	} // for
ret:
	return ndics > 0 ? self : nil;
}

- free
{
	int i;

	[self freeWordList];
	for (i = 0; i < ndics; i++)
		[dic_list[i] free];
	if (dic_list != NULL)
		free(dic_list);
	[super free];
	return nil;
}

- lookup:(char *)korean
{
	int i, j, k;
	int wc;
	
	[self freeWordList];
	
	for (wc = 0, i = 0; i < ndics; i++) {
		[dic_list[i] lookup:korean];
		wc += dic_list[i]->words_count;
	}

	if ((words = malloc(wc * sizeof(words[0]))) == NULL) {
		perror("ChineseDic");
		return nil;
	}
	for (k = 0, i = 0; i < ndics; i++)
		for (j = 0; j < dic_list[i]->words_count; j++, k++)
			words[k] = dic_list[i]->words[j];

	return (words_count = k) > 0 ? self : nil;
}

//
// private methods
//

- freeWordList
{
	words_count = 0;
	if (words != NULL) {
		free(words);
		words = NULL;
	}
	return self;
}

@end

@implementation PerFileDic

static char *entry_to_find;		// These 4 variables are accessed in -lookup:
static int entry_len;			// and compare_key_and_indexed_entry().
static char *dic;
static int dic_size;

- init
{
	[self free];
	return nil;
}

- initFromFile:(char *)path
{
	if ((self = [super init])) {
		char index_file[FILENAME_MAX];
		struct stat dst, ist;
		int dfd, ifd;
	
		cd = (void *)-1;
		cdi = (void *)-1;
	
		sprintf(index_file, "%si", path);

		// check if index is older than dictionary.

		if (stat(path, &dst) < 0) {
			perror(path);
			goto fail;
		}
		if (stat(index_file, &ist) < 0) {
			perror(index_file);
			goto fail;
		}
		if (ist.st_mtime < dst.st_mtime) { // index is older!
			fprintf(stderr, "PerFileDic: %s is older than %s.\n",
					index_file, path);
			goto fail;
		}

		if ((dfd = open(path, O_RDONLY)) < 0) {
			perror(path);
			goto fail;
		}
		cd = mmap(NULL, dst.st_size, PROT_READ, MAP_SHARED, dfd, 0);
		close(dfd);
		if (cd == (void *)-1) {
			perror(path);
			goto fail;
		}
		if ((ifd = open(index_file, O_RDONLY)) < 0) {
			perror(index_file);
			goto fail;
		}
		cdi = mmap(NULL, ist.st_size, PROT_READ, MAP_SHARED, ifd, 0);
		close(ifd);
		if (cdi == (void *)-1) {
			perror(index_file);
			goto fail;
		}
	
		cd_size = dst.st_size;
		cdi_nmemb = ist.st_size / sizeof(cdi[0]);

		return self;
	}
fail:
	[self free];
	return nil;
}

- free
{
	if (cd != (void *)-1)
		munmap(cd, cd_size);
	if (cdi != (void *)-1)
		munmap(cdi, cdi_nmemb * sizeof(cdi[0]));
	[self freeWordList];
	[super free];
	return nil;
}

- lookup:(char *)korean
{
	CDIndex *i;
	static int compare_key_and_indexed_entry(CDIndex *dummy, CDIndex *e);

	[self freeWordList];
	
	dic = cd;
	dic_size = cd_size;
	entry_to_find = korean;
	entry_len = strlen(entry_to_find);
	
	if ((i = bsearch(NULL, cdi, cdi_nmemb, sizeof(cdi[0]),
					 (BsearchComp)compare_key_and_indexed_entry)) != NULL)
	{
		char *p;
		char *eof = cd + cd_size;
		int n;
		
		// count words
		
		p = next_occurrence_of_space(cd + *i, eof);
		p = next_occurrence_of_nonspace(p, eof);
		for (n = 0; p < eof && !(p > cd && p[-1] == '\n');
			 p = next_occurrence_of_space(p, eof),
			 p = next_occurrence_of_nonspace(p, eof))
			n++;

		// store words
		
		if ((words = malloc(n * sizeof(words[0]))) == NULL) {
			perror("PerFileDic");
			return nil;
		}
		
		p = next_occurrence_of_space(cd + *i, eof);
		p = next_occurrence_of_nonspace(p, eof);
		for (words_count = 0; words_count < n; words_count++) {
			char *q = next_occurrence_of_space(p, eof);
			
			if ((words[words_count] = malloc(q - p + 1)) == NULL) {
				perror("PerFileDic");
				goto ret;
			}
			memcpy(words[words_count], p, q - p);
			words[words_count][q - p] = '\0';

			p = next_occurrence_of_nonspace(q, eof);
		}
	}
ret:
	return words_count > 0 ? self : nil;
}

static int compare_key_and_indexed_entry(CDIndex *dummy, CDIndex *e)
{
	char *dic_entry;
	int dic_entry_len;
	
	dic_entry = dic + *e;
	dic_entry_len = next_occurrence_of_space(dic_entry, dic + dic_size) -
		dic_entry;

	return memmemcmp(entry_to_find, entry_len, dic_entry, dic_entry_len);
}

//
//  private methods
//

- freeWordList
{
	int i;
	
	for (i = 0; i < words_count; i++)
		free(words[i]);
	words_count = 0;
	if (words != NULL) {
		free(words);
		words = NULL;
	}
	return self;
}

@end

static char *eod;				// accessed only in compare_dic_entries()
								// and sort_and_index().

static int count_entries(char *file, int file_size);
static BOOL construct_item_list(char **list, char *file, int size);
static BOOL write_sorted_dic_and_index(char *path, char **l, int n,
									   char *f, int size);
static int compare_dic_entries(char **e1, char **e2);

BOOL sort_and_index(char *path)
{
	int fd;
	char *cd = (char *)-1;
	int cd_size;
	char **item_list = NULL;
	int nmemb;
		
	if ((fd = open(path, O_RDONLY)) < 0) {
		perror(path);
		goto fail;
	}
	cd_size = lseek(fd, 0, SEEK_END);
	cd = mmap(NULL, cd_size, PROT_READ, MAP_SHARED, fd, 0);
	close(fd);
	if (cd == (void *)-1) {
		perror("sort_and_index()");
		goto fail;
	}

	nmemb = count_entries(cd, cd_size);
	if ((item_list = malloc(nmemb * sizeof(item_list[0]))) == NULL) {
		perror("sort_and_index()");
		goto fail;
	}
	if (!construct_item_list(item_list, cd, cd_size))
		goto fail;
	eod = cd + cd_size;
	qsort(item_list, nmemb, sizeof(item_list[0]),
		  (QsortComp)compare_dic_entries);
	if (!write_sorted_dic_and_index(path, item_list, nmemb, cd, cd_size))
		goto fail;

	free(item_list);
	munmap(cd, cd_size);
	return YES;

fail:
	if (item_list != NULL)
		free(item_list);
	if (cd != (void *)-1)
		munmap(cd, cd_size);
	return NO;
}


static int count_entries(char *file, int file_size)
{
	char *eof = file + file_size;
	char *p;
	int count;
	
	for (p = file, count = 0; p < eof; p = next_line(p, eof))
		if (!isspace(*p))
			count++;

	return count;
}

static BOOL construct_item_list(char **e, char *file, int size)
{
	char *eof = file + size;
	char *p;

	for (p = file; p < eof;)
		if (!isspace(*p)) {		// begin of an item
			*e++ = p;
			for (p = next_line(p, eof); p < eof && isspace(*p);)
				p = next_line(p, eof);
		}
		else
			p = next_line(p, eof);

	return YES;
}

static int compare_dic_entries(char **e1, char **e2)
{
	char *entry1_end, *entry2_end;
	int entry1_len, entry2_len;
	
	entry1_end = next_occurrence_of_space(*e1, eod);
	entry2_end = next_occurrence_of_space(*e2, eod);
	entry1_len = entry1_end - *e1;
	entry2_len = entry2_end - *e2;
	
	return memmemcmp(*e1, entry1_len, *e2, entry2_len);
}

static BOOL write_sorted_dic_and_index(char *path, char **list, int nmemb,
									   char *f, int size)
{
	int i;
	char *eof = f + size;
	char new_path[FILENAME_MAX];
	FILE *fp1 = NULL, *fp2 = NULL;
	
	sprintf(new_path, "%s.orig", path);	// backup file name
	if (rename(path, new_path) < 0) {
		perror(new_path);
		goto fail;
	}
	
	sprintf(new_path, "%si", path);	// index file name

	if ((fp1 = fopen(path, "w")) == NULL) {
		perror(path);
		goto fail;
	}
	if ((fp2 = fopen(new_path, "w")) == NULL) {
		perror(new_path);
		goto fail;
	}

	for (i = 0; i < nmemb; i++) {
		char *b, *e;
		
		if (i > 0 && compare_dic_entries(&list[i], &list[i-1]) == 0)
			// duplicacy
			b = next_occurrence_of_space(list[i], eof);
		else {
			CDIndex ind;
			
			b = list[i];
			ind = ftell(fp1);
			if (fwrite((void *)&ind, sizeof(ind), 1, fp2) != 1) {
				perror(new_path);
				goto fail;
			}
		}
		
		for (e = next_line(b, eof); e < eof && isspace(*e);)
			e = next_line(e, eof);

		if (fwrite(b, 1, e - b, fp1) != e - b) {
			perror(path);
			goto fail;
		}
		if (e == eof && e[-1] != '\n')	// file not terminated with '\n'.
			putc('\n', fp1);
	}

	fclose(fp1);
	fclose(fp2);
	return YES;

fail:
	if (fp1 != NULL)
		fclose(fp1);
	if (fp2 != NULL)
		fclose(fp2);
	return NO;
}
