// -*- mode: objc; tab-width: 4 -*-

//  This file is part of Byeoroo, a Korean input method server for X11.
//  Copyright (C) 1997  Park, Jae-hyon.
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License , or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//  
//  You may contact the author by:
//     e-mail:  jhpark@entropy.kaist.ac.kr
//       mail:  Park, Jae-hyon
//              Physics Department, 305-701
//              Korea Advanced Institute of Science and Technology
//              Taejon, South Korea


#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdlib.h>

#import  "KeyTranslator.h"
#import  "KAutomataForX.h"
#include "sutil.h"

#define MAX(a, b) ((a) > (b) ? (a) : (b))

typedef int (*QsortComp)(const void *, const void *);
typedef QsortComp BsearchComp;

@implementation KeyTranslator

static char *current_file;	// referenced only in -readKSM.

static BOOL prepare_buf(char **buf, int *buf_size, int required);

- init
{
	[self free];
	return nil;
}

- initFromFile:(char *)path
{
    int fd;
	
    tt = (void *)-1;
    
    if (!(self = [super init]))
		goto fail;

    if ((fd = open(path, O_RDONLY)) < 0) {
		perror("KeyTranslator");
		goto fail;
    }
    tt_size = lseek(fd, 0, SEEK_END);
	tt = mmap(NULL, tt_size, PROT_READ, MAP_SHARED, fd, 0);
	close(fd);
	if (tt == (void *)-1) {
		perror("KeyTranslator");
		goto fail;
    }

	current_file = path;
    if ([self readKSM])
		return self;

fail:
    [self free];
    return nil;
}

- free
{
    if (tt != (void *)-1)
		munmap(tt, tt_size);
    if (map != NULL)
		free(map);
    if (string_return != NULL)
		free(string_return);
    [super free];
    return nil;
}

- (char *)stringFromKeycode:(KeyCode)keycode modifier:(unsigned int)modifier
{
    struct _keycode_to_string_map *e;
    struct _keycode_to_string_map t = { keycode };
    static int compare_keycodes(struct _keycode_to_string_map *e1,
								struct _keycode_to_string_map *e2);

    if ((e = bsearch(&t, map, number, sizeof(map[0]),
					 (BsearchComp)compare_keycodes)) == NULL)
		return NULL;

    for (; e > map && e[-1].keycode == t.keycode; e--)
		;
    for (; e < map + number && e->keycode == t.keycode; e++)
		if ((modifier & e->modifier_mask) == e->modifier) {
			if (!prepare_buf(&string_return, &string_size, e->length+1))
				return NULL;
			else {
				memcpy(string_return, e->string, e->length);
				string_return[e->length] = '\0';
				return string_return;
			}
		}
    
    return NULL;
}

static int compare_keycodes(struct _keycode_to_string_map *e1,
							struct _keycode_to_string_map *e2)
{
    return (signed int)e1->keycode - (signed int)e2->keycode;
}

- (BOOL)translating
{
    return number > 0;
}

//
//  private methods
//

- readKSM
{
    char *eof = tt + tt_size;
    char *implicit_keys = tt, *implicit_keys_end = tt;
    int num_non_comment_lines = 0;
    char *p;
    int l;			// line number
    char *buf = NULL;
    int buf_size = 0;
    static int compare_map_items(struct _keycode_to_string_map *e1,
								 struct _keycode_to_string_map *e2);
    static int count_lines(char *p, char *eof);

    num_non_comment_lines = count_lines(tt, eof);
	// avoid malloc(0).
    if ((map = malloc(MAX(1, num_non_comment_lines * sizeof(map[0])))) == NULL)
	{
		perror("KeyTranslator");
		goto fail;
    }

    // parse key definititions.

    for (p = tt, l = 1;
		 p = eat_space(p, eof), p < eof;
		 p = next_line(p, eof), l++)
    {
		char *colon, *eol;
		KeySym keysym;
		KeyCode keycode;
		unsigned int modifier, modifier_mask;
		char *string;
		int length;

		if (*p == '!' || *p == '\n') // comment or blank line
			continue;

		eol = next_occurrence_of('\n', p, eof);
		if ((colon = next_occurrence_of(':', p, eol)) == eol) {	// no colon
			fprintf(stderr, "%s:%d: colon(`:') not found.\n", current_file, l);
			goto fail;
		}

		// treat implicit directives.

		if (memstrcmp(p, previous_occurrence_of_nonspace(colon, p) - p,
					  "Implicit") == 0)	// `Implicit:' directive
		{
			implicit_keys = colon + 1;
			implicit_keys_end = eol;
			continue;
		}

		// parse keycode and modifiers (left part of `:')

		if (!prepare_buf(&buf, &buf_size, (colon-p) +
						 (implicit_keys_end-implicit_keys) + 2))
			goto fail;
		memcpy(buf, p, colon - p);
		buf[colon - p] = ' ';
		strmemcpy(buf + (colon-p) + 1, implicit_keys,
				  implicit_keys_end - implicit_keys);

		if (!parse_key(&keysym, &modifier, &modifier_mask, buf)) {
			fprintf(stderr, "%s:%d: parse error.\n", current_file, l);
			goto fail;
		}
		if ((keycode = _keysym_to_keycode(keysym)) == 0) {
			fprintf(stderr, "%s:%d: cannot convert keysym `%s' to keycode.\n",
					current_file, l, XKeysymToString(keysym));
			goto fail;
		}

		// parse what to produce

		if ((p = next_occurrence_of('"', colon, eol)) == eol) {
			fprintf(stderr, "%s:%d: opening `\"' not found.\n", current_file,
					l);
			goto fail;
		}
		string = p + 1;
		if ((p = previous_occurrence_of('"', eol, string)) == string) {
			fprintf(stderr, "%s:%d: closing `\"' not found.\n", current_file,
					l);
			goto fail;
		}
		length = p-1 - string;

		map[number].keycode = keycode;
		map[number].modifier = modifier;
		map[number].modifier_mask = modifier_mask;
		map[number].string = string;
		map[number].length = length;
		number++;
    } // for

    map = realloc(map, number * sizeof(map[0]));
	if (number == 0)
		map = NULL;
	
    if (map != NULL)
		qsort(map, number, sizeof(map[0]), (QsortComp)compare_map_items);

    if (buf != NULL)
		free(buf);
    return self;

fail:
    if (map != NULL) {
		free(map);
		map = NULL;
    }
    if (buf != NULL)
		free(buf);
    return nil;
}

static int count_lines(char *p, char *eof)
{
    int n = 0;
	
    for (; p = eat_space(p, eof), p < eof; p = next_line(p, eof))
	if (*p != '!' || *p != '\n')
	    n++;

    return n;
}

static int compare_map_items(struct _keycode_to_string_map *e1,
							   struct _keycode_to_string_map *e2)
{
    if (e1->keycode != e2->keycode)
		return e1->keycode - e2->keycode;
    if (e1->modifier != e2->modifier)
		return (signed int)e1->modifier - (signed int)e2->modifier;
    if (e1->modifier_mask != e2->modifier_mask)
		return (signed int)e1->modifier_mask - (signed int)e2->modifier_mask;
    return 0;
}

static BOOL prepare_buf(char **buf, int *buf_size, int required)
{
    if (*buf == NULL || *buf_size < required) {
		char *new_buf;
			
		if ((new_buf = realloc(*buf, required)) == NULL) {
			perror("KeyTranslator");
			return NO;
		}

		*buf = new_buf;
		*buf_size = required;
    }

    return YES;
}

@end
