// -*- mode: objc; tab-width: 4 -*-

//  This file is part of Byeoroo, a Korean input method server for X11.
//  Copyright (C) 1997  Park, Jae-hyon.
//  
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License , or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//  
//  You may contact the author by:
//     e-mail:  jhpark@entropy.kaist.ac.kr
//       mail:  Park, Jae-hyon
//              Physics Department, 305-701
//              Korea Advanced Institute of Science and Technology
//              Taejon, South Korea


#import  "XChineseDic.h"

#include <stdlib.h>
#include <string.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/List.h>

#define MIN(a, b) ((a) < (b) ? (a) : (b))

@implementation XChineseDic

static void WordChosen(Widget w, id cl_d, XawListReturnStruct *c_d);
static void SetListWidth(Widget w);

- init
{
	return [self initFromSearchPath:"."];
}

- initFromSearchPath:(char *)search_path parentWidget:(Widget)w
{
	if ((self = [super initFromSearchPath:search_path])) {
		if (!(cdshell = XtVaCreatePopupShell("cdshell",
											 transientShellWidgetClass,
											 w, NULL)))
			goto fail;
		if (!(cdlist = XtVaCreateManagedWidget("cdlist", listWidgetClass,
											   cdshell, NULL)))
			goto fail;

		XtAddCallback(cdlist, XtNcallback, (XtCallbackProc)WordChosen, self);
		return self;
	}
fail:
	[self free];
	return nil;
}

- free
{
	if (cdshell)
		XtDestroyWidget(cdshell);
	if (cdlist)
		XtDestroyWidget(cdlist);
	if (list)
		free(list);
	if (entry)
		free(entry);
	[super free];
	return nil;
}

- lookup:(char *)korean
{
	if (entry)
		free(entry);
	if ((entry = strdup(korean)) == NULL)
		return nil;
	return [super lookup:korean];
}

- (Widget)getShell
{
	return cdshell;
}

- (Widget)getWidget
{
	return cdlist;
}

- popupCenteredAt:(Position)x :(Position)y
{
	if (words_count > 0) {
		Dimension   width, height;
		char **new_list;
		
		if ((new_list = realloc(list, (words_count+1) * sizeof(new_list[0])))
			== NULL)
			goto fail;
		list = new_list;
		memcpy(list, words, words_count * sizeof(words[0]));
		list[words_count] = entry;
		XawListChange(cdlist, list, words_count + 1, 0, True);
 		SetListWidth(cdlist);
		
		XtVaGetValues(cdshell, XtNwidth, &width, XtNheight, &height, NULL);

		x = x - width / 2;
		y = y - height / 2;

		if (x < 0)
			x = 0;
		else {
			int scr_width = WidthOfScreen(XtScreen(cdshell));

			if (x + width > scr_width)
				x = scr_width - width;
		}

		if (y < 0)
			y = 0;
		else {
			int scr_height = HeightOfScreen(XtScreen(cdshell));

			if (y + height > scr_height)
				y = scr_height - height;
		}

		XtVaSetValues(cdshell, XtNx, x, XtNy, y, NULL);

		XtPopup(cdshell, XtGrabExclusive);
		if (transient_for)
			XSetTransientForHint(XtDisplay(cdshell), XtWindow(cdshell),
								 transient_for);
		popped_up = YES;
		return self;
	}
fail:
	return nil;
}

- (BOOL)isPoppedup
{
	return popped_up;
}

- popdown
{
	XtPopdown(cdshell);
	popped_up = NO;
	return self;
}

- setTarget:(id)anObj
{
	target = anObj;
	return self;
}

- setAction:(SEL)sel
{
	action = sel;
	return self;
}

- setTransientFor:(Window)win
{
	transient_for = win;
	return self;
}

//
//  private methods
//

- wordChosen:(XtPointer)list_return
{
	XawListReturnStruct *list_ret = list_return;

	[self popdown];
	if (list_ret->list_index != words_count)
		[target perform:action with:(id)list_ret->string];
	return self;
}

static void WordChosen(Widget w, id self, XawListReturnStruct *list_ret)
{
	[self wordChosen:list_ret];
}

static void SetListWidth(Widget w)
{
	int longest;
	// Type, `int' is important!  If `longest' were declared as a Dimension,
	// XtVaGetValues() would corrupt the stack!
	int number;
	Dimension column_spacing;
	Dimension internal_width;
	Dimension one_row_width;
	Dimension half_screen_width;
	
	XtVaGetValues(w, XtNlongest, &longest, XtNnumberStrings, &number,
				  XtNcolumnSpacing, &column_spacing,
				  XtNinternalWidth, &internal_width, NULL);
	one_row_width = (longest + column_spacing) * number + 2 * internal_width;
	half_screen_width = WidthOfScreen(XtScreen(w)) / 2;
	XtVaSetValues(w, XtNwidth, MIN(one_row_width, half_screen_width), NULL);
}

@end
